import React, { memo, useRef, useState } from 'react';

import { useTheme } from '@mui/material';
import { Handle, Position } from 'reactflow';

import styles from './relationpill.module.scss';
import { RelationNode } from '../../../graph/reactflow/model';
import { Handles } from '../../../graph/reactflow/handles';

// export type RelationRFPillProps = {
//   data: {
//     name: string;
//     suggestedForConnection: any;
//     isFromEntityConnected?: boolean;
//     isToEntityConnected?: boolean;
//   };
// };

/**
 * Component to render a relation flow element
 * @param { FlowElement<RelationData>} param0 The data of a relation flow element.
 */
export const RelationPill = memo(({ data }: RelationNode) => {
  // export default function RelationRFPill({ data }: { data: any }) {
  const theme = useTheme();
  // console.log('RelationRFPill', data);

  const minRef = useRef<HTMLInputElement>(null);
  const maxRef = useRef<HTMLInputElement>(null);

  const [readOnlyMin, setReadOnlyMin] = useState(true);
  const [readOnlyMax, setReadOnlyMax] = useState(true);

  const onDepthChanged = (depth: string) => {
    // Don't allow depth above 99
    const limit = 99;
    if (data != undefined) {
      data.depth.min = data.depth.min >= limit ? limit : data.depth.min;
      data.depth.max = data.depth.max >= limit ? limit : data.depth.max;

      // Check for for valid depth: min <= max
      if (depth == 'min') {
        if (data.depth.min > data.depth.max) data.depth.max = data.depth.min;
        setReadOnlyMin(true);
      } else if (depth == 'max') {
        if (data.depth.max < data.depth.min) data.depth.min = data.depth.max;
        setReadOnlyMax(true);
      }

      // Set to the correct width
      if (maxRef.current)
        maxRef.current.style.maxWidth = calcWidth(data.depth.max);
      if (minRef.current)
        minRef.current.style.maxWidth = calcWidth(data.depth.min);
    }
  };

  const isNumber = (x: string) => {
    {
      if (typeof x != 'string') return false;
      return !Number.isNaN(x) && !Number.isNaN(parseFloat(x));
    }
  };

  const calcWidth = (data: number) => {
    return data.toString().length + 0.5 + 'ch';
  };

  return (
    <div
      className={styles.relation}
      style={{ backgroundColor: theme.palette.custom.elements.relation[0] }}
    >
      <div
        className={styles.rightArrow}
        style={{ borderLeftColor: theme.palette.custom.elements.relation[0] }}
      ></div>
      <div
        className={styles.leftArrow}
        style={{ borderRightColor: theme.palette.custom.elements.relation[0] }}
      ></div>
      {/* <span
        className={styles.relationTop}
        style={{ backgroundColor: theme.palette.custom.elements.relation[0] }}
      ></span>
      <span
        className={styles.relationBottom}
        style={{ backgroundColor: theme.palette.custom.elements.relation[0] }}
      ></span> */}
      <div className={styles.relationWrapper}>
        <span
          className={styles.relationHandleFiller}
          // style={{ transform: 'translate(-100px,0)' }}
        >
          <Handle
            id={Handles.RelationLeft}
            type="target"
            position={Position.Left}
            className={
              styles.relationHandleLeft +
              ' ' +
              (false ? styles.handleConnectedBorderLeft : '')
            }
          />
        </span>
        <span className={styles.relationHandleFiller}>
          <Handle
            id={Handles.ToAttribute}
            type="target"
            position={Position.Left}
            className={
              styles.relationHandleAttribute +
              ' ' +
              (false ? styles.handleConnectedFill : '')
            }
          />
        </span>
        <span className={styles.relationHandleFiller}>
          <Handle
            id={Handles.ReceiveFunction}
            type="target"
            position={Position.Left}
            className={
              styles.relationHandleFunction +
              ' ' +
              (false ? styles.handleConnectedFill : '')
            }
          />
        </span>
        <div className={styles.relationDataWrapper}>
          <span className={styles.relationSpan}>{data?.name}</span>
          <span className={styles.relationInputHolder}>
            <span>[</span>
            <input
              className={
                styles.relationInput +
                ' ' +
                (readOnlyMin ? styles.relationReadonly : '')
              }
              ref={minRef}
              type="string"
              min={0}
              readOnly={readOnlyMin}
              placeholder={'?'}
              value={data?.depth.min}
              onChange={(e) => {
                if (data != undefined) {
                  data.depth.min = isNumber(e.target.value)
                    ? parseInt(e.target.value)
                    : 0;
                  e.target.style.maxWidth = calcWidth(data.depth.min);
                }
              }}
              onDoubleClick={() => {
                setReadOnlyMin(false);
              }}
              onBlur={(e) => {
                onDepthChanged('min');
              }}
              onKeyDown={(e) => {
                if (e.key === 'Enter') {
                  onDepthChanged('min');
                }
              }}
            ></input>
            <span>..</span>
            <input
              className={
                styles.relationInput +
                ' ' +
                (readOnlyMax ? styles.relationReadonly : '')
              }
              ref={maxRef}
              type="string"
              min={0}
              readOnly={readOnlyMax}
              placeholder={'?'}
              value={data?.depth.max}
              onChange={(e) => {
                if (data != undefined) {
                  data.depth.max = isNumber(e.target.value)
                    ? parseInt(e.target.value)
                    : 0;
                  e.target.style.maxWidth = calcWidth(data.depth.max);
                }
              }}
              onDoubleClick={() => {
                setReadOnlyMax(false);
              }}
              onBlur={(e) => {
                onDepthChanged('max');
              }}
              onKeyDown={(e) => {
                if (e.key === 'Enter') {
                  onDepthChanged('max');
                }
              }}
            ></input>
            <span>]</span>
          </span>
        </div>
        <span className={styles.relationHandleFiller}>
          <Handle
            id={Handles.RelationRight}
            type="target"
            position={Position.Right}
            className={
              styles.relationHandleRight +
              ' ' +
              (false ? styles.handleConnectedBorderRight : '')
            }
          />
        </span>
      </div>
    </div>
  );
});

RelationPill.displayName = 'RelationPill';
export default RelationPill;
